from __future__ import print_function
from __future__ import unicode_literals

import boto_source_profile_mfa

from botocore.exceptions import ClientError, ParamValidationError

from functools import wraps

from . import config


_sessions = {}


def retry(attempts=3):
    def wrapper(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            tries = attempts
            while True:
                tries -= 1
                try:
                    return func(*args, **kwargs)
                except (ClientError, ParamValidationError) as error:
                    if tries > 0:
                        print('[ssha] {}'.format(error))
                    else:
                        raise
        return wrapped
    return wrapper


@retry()
def client(*args, **kwargs):
    return session().client(*args, **kwargs)


@retry()
def credentials(*args, **kwargs):
    return session().get_credentials(*args, **kwargs)


@retry()
def resource(*args, **kwargs):
    return session().resource(*args, **kwargs)


def session():
    aws_config = config.get('aws') or {}
    return boto_source_profile_mfa.get_session(**aws_config)



from __future__ import print_function
from __future__ import unicode_literals

import errno
import os

from . import config


def _get_address(instance_ip):

    username = config.get('ssh.username_prefix', '') + config.get('ssh.username', '')

    # Don't add the username to the address when it is the current user,
    # because it would make no difference.
    if username == os.environ.get('USER'):
        username = None

    if username:
        return username + '@' + instance_ip
    else:
        return instance_ip


def connect(instance, bastion, command):

    bastion_hostname = config.get('bastion.hostname')
    if not bastion_hostname and bastion:
        bastion_hostname = get_ip(bastion, connect_through_bastion=False)
    if bastion_hostname:
        config.add('bastion.address', _get_address(bastion_hostname))

    instance_ip = get_ip(instance, connect_through_bastion=bool(bastion_hostname))
    config.add('hostname', instance_ip)

    instance_address = _get_address(instance_ip)
    config.add('address', instance_address)

    ssh_command = ['ssh']
    if config.get('verbose'):
        ssh_command += ['-v']
    user_known_hosts_file = config.get('ssh.user_known_hosts_file')
    if user_known_hosts_file:
        ssh_command += ['-o', 'UserKnownHostsFile={}'.format(user_known_hosts_file)]
    if bastion_hostname:
        proxy_command = config.get('ssh.proxy_command')
        ssh_command += ['-o', 'ProxyCommand={}'.format(proxy_command)]
    ssh_command += [instance_address]
    config.add('ssh.cmd', format_command(ssh_command))

    if command:

        command = config.render(command)
        print('[ssha] running {}'.format(command))
        return os.system(command)

    else:

        print('[ssha] running {}'.format(config.get('ssh.cmd')))
        run(ssh_command)


def format_command(command):
    args = []
    for arg in command:
        if ' ' in arg:
            args.append('"' + arg + '"')
        else:
            args.append(arg)
    return ' '.join(args)


def get_ip(instance, connect_through_bastion):
    if connect_through_bastion:
        return instance['PrivateIpAddress']
    return instance.get('PublicIpAddress') or instance['PrivateIpAddress']


def run(command):
    child_pid = os.fork()
    if child_pid == 0:
        os.execlp(command[0], *command)
    else:
        while True:
            try:
                os.waitpid(child_pid, 0)
            except OSError as error:
                if error.errno == errno.ECHILD:
                    # No child processes.
                    # It has exited already.
                    break
                elif error.errno == errno.EINTR:
                    # Interrupted system call.
                    # This happens when resizing the terminal.
                    pass
                else:
                    # An actual error occurred.
                    raise



import collections
import curses
import re

from curses import panel

from . import ec2


Item = collections.namedtuple('Item', field_names=('label', 'value'))


class Menu(object):

    def __init__(self, title, items, stdscreen):
        self.window = stdscreen.subwin(0, 0)
        self.window.timeout(1000)
        self.window.keypad(1)
        self.panel = panel.new_panel(self.window)
        self.panel.hide()
        panel.update_panels()

        self.title = title

        self.position = 0
        self.items = items

        self.offset_x = 2

    def addstr(self, y, x, string, attr):
        try:
            self.window.addstr(y, x, string, attr)
        except curses.error:
            # Curses will error on the last line even when it works.
            # https://stackoverflow.com/questions/7063128/last-character-of-a-window-in-python-curses
            if y == self.max_y - 1:
                pass
            else:
                raise

    def addline(self, y, string, attr):
        """
        Displays a string on the screen. Handles truncation and borders.

        """

        if y >= self.max_y:
            return

        # Display the left blank border.
        self.addstr(
            y=y,
            x=0,
            string=' ' * self.offset_x,
            attr=curses.A_NORMAL,
        )

        # Remove trailing spaces so the truncate logic works correctly.
        string = string.rstrip()

        # Truncate the string if it is too long.
        if self.offset_x + len(string) + self.offset_x > self.max_x:
            string = string[:self.max_x - self.offset_x - self.offset_x - 2] + '..'

        # Add whitespace between the end of the string and the edge of the
        # screen. This is required when scrolling, to blank out characters
        # from other lines that had been displayed here previously.
        string += ' ' * (self.max_x - self.offset_x - len(string) - self.offset_x)

        # Display the string.
        self.addstr(
            y=y,
            x=self.offset_x,
            string=string,
            attr=attr,
        )

        # Display the right blank border.
        self.addstr(
            y=y,
            x=self.max_x - self.offset_x,
            string=' ' * self.offset_x,
            attr=curses.A_NORMAL,
        )

    def navigate(self, n):
        self.position += n
        if self.position < 0:
            self.position = len(self.items) - 1
        elif self.position >= len(self.items):
            self.position = 0

    def display(self):

        self.panel.top()
        self.panel.show()
        self.window.clear()

        while True:

            self.window.refresh()
            curses.doupdate()

            self.max_y, self.max_x = self.window.getmaxyx()

            # Display the menu title.
            if self.title:
                self.addline(1, self.title)
                self.addline(2, '-' * len(self.title))
                offset_top = 3
            else:
                offset_top = 1
            offset_bottom = 1

            window_height = max(self.max_y - offset_top - offset_bottom - 1, 0)

            if self.position < window_height:
                window = (0, window_height)
            else:
                window = (self.position - window_height, self.position)

            row = 0

            for index, item in enumerate(self.items):

                if index < window[0] or index > window[1]:
                    continue

                # Highlight the selected item.
                if index == self.position:
                    mode = curses.A_REVERSE
                else:
                    mode = curses.A_NORMAL

                # Display the item.
                self.addline(
                    offset_top + row,
                    item.label,
                    mode,
                )

                row += 1

            # Blank bottom lines if screen was resized
            for y in range(offset_bottom):
                self.addline(
                    self.max_y - y - 1,
                    '',
                    curses.A_NORMAL,
                )

            # Because window.timeout was called,
            # this returns -1 if nothing was pressed.
            key = self.window.getch()

            if key in [curses.KEY_ENTER, ord('\n')]:
                return self.items[self.position].value
            elif key in (curses.KEY_UP, ord('k')):
                self.navigate(-1)
            elif key in (curses.KEY_DOWN, ord('j')):
                self.navigate(1)
            elif key in (ord('q'), ord('Q')):
                raise KeyboardInterrupt

        self.window.clear()
        self.panel.hide()
        panel.update_panels()
        curses.doupdate()


def _display(screen, items, title=None):
    curses.curs_set(0)
    menu = Menu(title, items, screen)
    return menu.display()


def choose_config(names, search):
    if search:
        return search
    elif len(names) > 1:
        items = [Item(label=name, value=name) for name in names]
        return curses.wrapper(_display, items)
    elif names:
        return names[0]
    else:
        return None


def _find_each_column_width(table):
    if not table:
        return []

    columns_size = [0] * len(table[0])
    for row in table:
        for j, column_element in enumerate(row):
            columns_size[j] = max(columns_size[j], len(column_element))
    return columns_size


def choose_instance(instances, search, show_menu=True):

    labels = [ec2.label(inst) for inst in instances]
    columns_width = _find_each_column_width(labels)

    items = []
    for i, inst in enumerate(instances):
        formatted_labels = [label.ljust(columns_width[j]) for j, label in enumerate(labels[i])]
        items.append(Item(label=' '.join(formatted_labels), value=inst))

    if search:
        search = search.lower()
        items = [item for item in items if search in item.label.lower()]
        if len(items) == 1:
            return items[0].value

    if not items:
        return None

    if not show_menu:
        if search:
            # If there is a search string, use word boundaries
            # to increase the changes of it being a good choice.
            pattern = r'\b{}\b'.format(re.escape(search))
            items.sort(
                key=lambda item: bool(re.search(pattern, item.label, re.IGNORECASE)),
                reverse=True,
            )
        return items[0].value

    return curses.wrapper(_display, items)



from boto_source_profile_mfa import SourceProfileMfaCredentialProvider

from botocore.credentials import AssumeRoleProvider

from . import aws, errors


_cache = {}


def _list_groups_for_user(username):

    iam = aws.client('iam')

    print('[ssha] discovering iam groups')

    groups = []

    marker = None
    while True:

        kwargs = {
            'UserName': username,
        }
        if marker:
            kwargs['Marker'] = marker

        response = iam.list_groups_for_user(**kwargs)
        if response['ResponseMetadata']['HTTPStatusCode'] != 200:
            errors.json_exit(response)

        for group in response['Groups']:
            groups.append(group['GroupName'])

        marker = response.get('Marker')
        if not marker:
            return groups


def groups():
    if 'groups' not in _cache:
        username = user()
        if username:
            _cache['groups'] = _list_groups_for_user(username)
        else:
            _cache['groups'] = []
    return _cache['groups']


def user():
    if 'user' not in _cache:
        creds = aws.credentials()
        if creds.method in (SourceProfileMfaCredentialProvider.METHOD, AssumeRoleProvider.METHOD):
            _cache['user'] = None
        else:
            print('[ssha] discovering iam user')
            iam = aws.resource('iam')
            user = iam.CurrentUser()
            _cache['user'] = user.user_name
    return _cache['user']



#!/usr/bin/env python

from setuptools import setup

import ssha


setup(
    name='ssha',
    version=ssha.__version__,
    description='SSH into AWS EC2 instances',
    author='Raymond Butcher',
    author_email='ray.butcher@claranet.uk',
    url='https://github.com/claranet/ssha',
    license='MIT License',
    packages=(
        'ssha',
    ),
    scripts=(
        'bin/ssha',
    ),
    install_requires=(
        'boto-source-profile-mfa>=0.0.8',
        'botocore>=1.5.8',
        'boto3',
        'cryptography',
        'paramiko',
        'pyhcl',
    ),
)
